/*
   interpreter.cpp - version 1.11, 2002

   cInterpreter is a class that interprets a character string
   with a mathematical expression.

   Copyright (c) 2001-2002 Andreas Jnsson

   This software is provided 'as-is', without any form of
   warranty. In no case will the author be held responsible
   for any damage caused by its use.

   Permission is granted to anyone to use the software for
   for any purpose, including commercial. It is also allowed
   to modify the software and redistribute it free of charge.
   The permission is granted with the following restrictions:

   1. The origin of the software may not be misrepresented.
      It must be plainly understandable who is the author of
      the original software.
   2. Altered versions must not be misrepresented as the
      original software, i.e. must be plainly marked as
      altered.
   3. This notice may not be removed or altered from any
      distribution of the software, altered or not.

   Andreas Jnsson
   andreas@angelcode.com
*/

#include "interpreter.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

//=============================================================================

cInterpreter::cInterpreter()
{
	lastAnswer = 0;
}

//=============================================================================

// Interprets a mathematical expression and returns the result in a string
void cInterpreter::InterpretLine(const char *line, char *output)
{
	double result;

	try
	{
		line = EatWhites(line);
		line = GetExpression(line, &result); 
		line = EatWhites(line);
		if( *line != '\0' )
		{
			cCalcException *e = new cCalcException("Unexpected character!\n");
			throw e;
		}
	}
	catch( cCalcException *e )
	{
		sprintf(output, e->message);

		delete e;
		return;
	}

	lastAnswer = result;
	sprintf(output, "= %.15g\n", result);
}

//=============================================================================

char *cInterpreter::EatWhites(const char *line)
{
	while(*line == ' ' || *line == '\t')
		line++;

	return (char *)line;
}

//=============================================================================

char *cInterpreter::GetExpression(const char *line, double *pResult)
{
	double term1, term2;
	
	term1 = 0;
	term2 = 2;

	line = GetTerm(line, &term1);

	while(1)
	{
		line = EatWhites(line);

		if( *line == '-' )
		{
			line++;
			line = EatWhites(line);
			line = GetTerm(line, &term2);
			term1 -= term2;
		}
		else if( *line == '+' )
		{
			line++;
			line = EatWhites(line);
			line = GetTerm(line, &term2);
			term1 += term2;
		}
		else
			break;
	}

	*pResult = term1;

	return (char *)line;
}

//=============================================================================

char *cInterpreter::GetTerm(const char *line, double *pResult)
{
	double factor1, factor2;

	factor1 = 1;
	factor2 = 0;

	line = GetFactor(line, &factor1);

	while(1)
	{
		line = EatWhites(line);

		if( *line == '*' )
		{
			line++;
			line = EatWhites(line);
			line = GetFactor(line, &factor2);
			factor1 *= factor2;
		}
		else if( *line == '/' )
		{
			line++;
			line = EatWhites(line);
			line = GetFactor(line, &factor2);
			if( factor2 < 1.0e-25 )
			{
				cCalcException *e = new cCalcException("Division by zero!\n");
				throw e;
			}

			factor1 /= factor2;
		}
		else
			break;
	}

	*pResult = factor1;

	return (char *)line;
}

//=============================================================================

char *cInterpreter::GetFactor(const char *line, double *pResult)
{
	if( *line == '-' )
	{
		line++;
		line = GetPower(line, pResult);
		*pResult = -(*pResult);
		return (char *)line;
	}
	
	return GetPower(line, pResult);
}

//=============================================================================

char *cInterpreter::GetPower(const char *line, double *pResult)
{
	double number1, number2;

	line = GetNumber(line, &number1);

	if( *line == '^' )
	{
		line++;
		line = GetNumber(line, &number2);
		*pResult = pow(number1,number2);
		return (char *)line;
	}

	*pResult = number1;
	return (char *)line;
}

//=============================================================================

char *cInterpreter::GetNumber(const char *line, double *pResult)
{
	double number = 0;

	if( *line == '\0' )
	{
		*pResult = 0;
		return (char *)line;
	}	

	if( *line == '(' )
	{
		line++;
		line = GetExpression(line, pResult);

		line = EatWhites(line);
		if( *line != ')' )
		{
			cCalcException *e = new cCalcException("Missing ')' in expression!\n");
			throw e;
		}

		return (char *)++line;
	}

	if( *line >= 'a' && *line <= 'z' || *line >= 'A' && *line <= 'Z' || *line == '_' )
	{
		return GetFunction(line, pResult);
	}

	// [digits][.digits][{e|E}[+|-]digits]
	if( *line >= '0' && *line <= '9' || *line == '.' )
	{
		number = atof(line);
		*pResult = number;

		// Read the number
		while( *line >= '0' && *line <= '9' )
		{
			line++;
		}

		if( *line == '.' )
		{
			line++;
			while( *line >= '0' && *line <= '9' )
				line++;
		}

		if( *line == 'e' || *line == 'E' )
		{
			line++;
			if( *line == '-' || *line == '+' )
				line++;

			while( *line >= '0' && *line <= '9' )
				line++;
		}
	}
	else
	{
		cCalcException *e = new cCalcException("Not a valid number!\n");
		throw e;
	}

	return (char *)line;
}

//=============================================================================

char *cInterpreter::GetFunction(const char *line, double *pResult)
{
	char name[256];
	
	line = GetName(line, name);

	//-------------------------------------------------------------------------
	// Constants
	if( strcmp(name, "ans") == 0 )
		*pResult = lastAnswer;
	else if( strcmp(name, "pi") == 0 )
		*pResult = 3.1415926535897931;
	else if( strcmp(name, "e") == 0 )
		*pResult = 2.7182818284590451;
	else if( strcmp(name, "d2r") == 0 )
		*pResult = 0.017453292519943295;
	else if( strcmp(name, "r2d") == 0 )
		*pResult = 57.295779513082323;

	//-------------------------------------------------------------------------
	// Trigonometry
	else if( strcmp(name, "sin") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = sin(number);
	}
	else if( strcmp(name, "cos") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = cos(number);
	}
	else if( strcmp(name, "tan") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = tan(number);
	}
	else if( strcmp(name, "asin") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);

		if( number > 1 || number < -1 )
		{
			cCalcException *e = new cCalcException("Argument out of range!\n");
			throw e;
		}

		*pResult = asin(number);
	}
	else if( strcmp(name, "acos") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);

		if( number > 1 || number < -1 )
		{
			cCalcException *e = new cCalcException("Argument out of range!\n");
			throw e;
		}

		*pResult = acos(number);
	}
	else if( strcmp(name, "atan") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = atan(number);
	}

	//-------------------------------------------------------------------------
	// Hyperbolics
	else if( strcmp(name, "sinh") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = sinh(number);
	}
	else if( strcmp(name, "cosh") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = cosh(number);
	}
	else if( strcmp(name, "tanh") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = tanh(number);
	}
		
	//-------------------------------------------------------------------------
	// Integers
	else if( strcmp(name, "abs") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = fabs(number);
	}
	else if( strcmp(name, "ceil") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = ceil(number);
	}
	else if( strcmp(name, "floor") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = floor(number);
	}
	else if( strcmp(name, "mod") == 0 )
	{
		double number[2];
		line = GetArguments(line, 2, number);
	
		if( number[1] < 1.0e-25 )
		{
			cCalcException *e = new cCalcException("Division by zero!\n");
			throw e;
		}

		*pResult = fmod(number[0],number[1]);
	}
	else if( strcmp(name, "div") == 0 )
	{
		double number[2];
		line = GetArguments(line, 2, number);

		if( number[1] < 1.0e-25 )
		{
			cCalcException *e = new cCalcException("Division by zero!\n");
			throw e;
		}

		double div = number[0]/number[1];
		modf(div, pResult);
	}
	else if( strcmp(name, "frac") == 0 )
	{
		double number, integer;
		line = GetArguments(line, 1, &number);
		*pResult = modf(number, &integer);
	}
	else if( strcmp(name, "int") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		modf(number, pResult);
	}

	//-------------------------------------------------------------------------
	// Logarithms
	else if( strcmp(name, "exp") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);
		*pResult = exp(number);
	}
	else if( strcmp(name, "log") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);

		if( number < 0 )
		{
			cCalcException *e = new cCalcException("Argument out of range!\n");
			throw e;
		}
		
		*pResult = log(number);
	}
	else if( strcmp(name, "log10") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);

		if( number < 0 )
		{
			cCalcException *e = new cCalcException("Argument out of range!\n");
			throw e;
		}

		*pResult = log10(number);
	}

	//-------------------------------------------------------------------------
	// Power functions
	else if( strcmp(name, "sqrt") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);

		if( number < 0 )
		{
			cCalcException *e = new cCalcException("Argument out of range!\n");
			throw e;
		}

		*pResult = sqrt(number);
	}

	//-------------------------------------------------------------------------
	// Memory functions
	else if( strcmp(name, "store") == 0 )
	{
		double numbers[2];
		line = GetArguments(line, 2, numbers);

		if( numbers[0] < 0 )
		{
			cCalcException *e = new cCalcException("Memory id must be positive\n");
			throw e;
		}
		if( numbers[0] != int(numbers[0]) )
		{
			cCalcException *e = new cCalcException("Memory id must be integer\n");
			throw e;
		}

		// Store the value in memory
		*pResult = memory.Store(int(numbers[0]), numbers[1]);
	}
	else if( strcmp(name, "recall") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);

		if( number < 0 )
		{
			cCalcException *e = new cCalcException("Memory id must be positive\n");
			throw e;
		}
		if( number != int(number) )
		{
			cCalcException *e = new cCalcException("Memory id must be integer\n");
			throw e;
		}

		// Recall the value from memory
		*pResult = memory.Recall(int(number));
	}
	else if( strcmp(name, "clear") == 0 )
	{
		double number;
		line = GetArguments(line, 1, &number);

		if( number < 0 )
		{
			cCalcException *e = new cCalcException("Memory id must be positive\n");
			throw e;
		}
		if( number != int(number) )
		{
			cCalcException *e = new cCalcException("Memory id must be integer\n");
			throw e;
		}

		// Recall and remove the value from memory
		*pResult = memory.Clear(int(number));
	}

	//-------------------------------------------------------------------------
	else
	{
		cCalcException *e = new cCalcException("Unrecognized function!\n");
		throw e;
	}

	return (char *)line;
}

//=============================================================================

char *cInterpreter::GetName(const char *line, char *name)
{
	int n = 0;
	while( *line >= 'a' && *line <= 'z' || 
		   *line >= 'A' && *line <= 'Z' || *line == '_' ||
		   *line >= '0' && *line <= '9' )
	{
		*name++ = *line++;
		if( ++n > 256 )
		{
			cCalcException *e = new cCalcException("Function name is too long!\n");
			throw e;
		}
	}

	*name = '\0';
	return (char *)line;
}

//=============================================================================

char *cInterpreter::GetArguments(const char *line, int numArgs, double *pArgs)
{
	int n = 0;

	// Read the argument expression
	if( *line != '(' )
	{
		cCalcException *e = new cCalcException("Missing paranthesis '('!\n");
		throw e;
	}

	line++;
	line = EatWhites(line);

	line = GetExpression(line, &pArgs[n++]);
	while( n < numArgs )
	{
		line = EatWhites(line);
		if( *line != ',' )
		{
			cCalcException *e = new cCalcException("Expected comma ','!\n");
			throw e;
		}

		line++;
		line = EatWhites(line);

		line = GetExpression(line, &pArgs[n++]);
	}

	line = EatWhites(line);
	if( *line != ')' )
	{
		cCalcException *e = new cCalcException("Missing paranthesis ')'!\n");
		throw e;
	}

	line++;
	return (char *)line;
}